#!/usr/bin/env python
#**************************************************************************************
#**************************************************************************************
#  NIDM2BIDSMRI.py
#  License: Apache License, Version 2.0
#**************************************************************************************
#**************************************************************************************
# Date: 10-2-17                 Coded by: David Keator (dbkeator@gmail.com)
# Filename: NIDM2BIDSMRI.py
#
# Program description:  This program will convert a NIDM-Experiment RDF document
# to a BIDS dataset.  The program will query the NIDM-Experiment document for subjects,
# MRI scans, and associated assessments saving the MRI data to disk in an organization
# according to the BIDS specification, the demographics metadata to a participants.tsv
# file, the project-level metdata to a dataset_description.json file, and the
# assessments to *.tsv/*.json file pairs in a phenotypes directory.
#**************************************************************************************
# Development environment: Python - PyCharm IDE
#
#**************************************************************************************
# System requirements:  Python 3.X
# Libraries: pybids, numpy, matplotlib, pandas, scipy, math, dateutil, datetime,argparse,
# os,sys,getopt,csv
#**************************************************************************************
# Start date: 10-2-17
# Update history:
# DATE            MODIFICATION				Who
#
#
#**************************************************************************************
# Programmer comments:
#
#
#**************************************************************************************
#**************************************************************************************

import sys, getopt, os
from os.path import join, isfile, basename, isdir,splitext
from os import mkdir
from os import system

from nidm.experiment import Project,Session,Acquisition,AcquisitionObject,DemographicsObject,AssessmentObject, MRObject
from nidm.core import BIDS_Constants,Constants
from prov.model import PROV_LABEL,PROV_TYPE
from nidm.experiment.Utils import read_nidm
from nidm.experiment.Query import GetProjectsUUID, GetProjectLocation, GetParticipantIDFromAcquisition

import json
from pprint import pprint
import csv
import glob
from rdflib import Graph,URIRef,RDF
from argparse import ArgumentParser
from io import StringIO
import pandas as pd
import validators
import urllib.parse
from shutil import copyfile, move
import urllib.request as ur
import tempfile
import datalad.api as dl

def GetImageFromAWS(location, output_file,args):
    '''
    This function will attempt to get a BIDS image identified by location from AWS S3.  It only
    supports known URLs at this time (e.g. openneuro)
    :param location: path string to file. This can be a local path. Function will try and detect if this
    is a known project/archive and if so will format theh S3 string appropriately.  Otherwise it will return None
    :param output_file: This is the full path and filename to store the S3 downloaded file if successful
    :return: None if file not downloaded else will return True
    '''

    print("Trying AWS S3 for dataset: %s" % location)
    # modify location to remove everything before the dataset name
    # problem is we don't know the dataset identifier inside the path string because
    # it doesn't have any constraints.  For openneuro datasets they start with "ds" so
    # we could pick that out but for others it's difficult (impossible)?

    # case for openneuro
    if 'openneuro' in location:
        # remove everything from location string before openneuro
        openneuro_loc = location[location.find("openneuro/") + 10:]
        # get a temporary directory for this file
        temp_dir = tempfile.TemporaryDirectory()
        # aws command
        cmd = "aws s3 cp --no-sign-request " + "s3://openneuro.org/" + openneuro_loc + " " + temp_dir.name
        # execute command
        print(cmd)
        system(cmd)
        # check if aws command downloaded something
        if not isfile(join(temp_dir.name, basename(location))):
            print("Couldn't get dataset from AWS either...")
            return None
        else:
            try:
                # copy file from temp_dir to bids dataset
                print("Copying temporary file to final location....")
                copyfile(join(temp_dir.name, basename(location)),output_file)
                return True
            except:
                print("Couldn't get dataset from AWS either...")
                return None
    # if user supplied a URL base, add dataset, subject, and file information to it and try to download the image
    elif args.aws_baseurl:
        aws_baseurl = args.aws_baseurl
        # check if user supplied the last '/' in the aws_baseurl or not.  If not, add it.
        if aws_baseurl[-1] != '/':
            aws_baseurl = aws_baseurl = '/'
        # remove everything from location string before openneuro
        loc = location[location.find(args.dataset_string) + len(args.dataset_string):]
        # get a temporary directory for this file
        temp_dir = tempfile.TemporaryDirectory()
        # aws command
        cmd = "aws s3 cp --no-sign-request " + aws_baseurl + loc + " " + temp_dir.name
        # execute command
        print(cmd)
        system(cmd)
        # check if aws command downloaded something
        if not isfile(join(temp_dir.name, basename(location))):
            print("Couldn't get dataset from AWS either...")
            return None
        else:
            try:
                # copy file from temp_dir to bids dataset
                print("Copying temporary file to final location....")
                copyfile(join(temp_dir.name, basename(location)), output_file)
                return True
            except:
                print("Couldn't get dataset from AWS either...")
                return None



def GetImageFromURL(url):
    '''
    This function will try and retrieve the file referenced by url
    :param url: url to file to download
    :return: temporary filename or -1 if fails
    '''


    # try to open the url and get the pointed to file
    try:
        # open url and get file
        opener = ur.urlopen(url)
        # write temporary file to disk and use for stats
        temp = tempfile.NamedTemporaryFile(delete=False)
        temp.write(opener.read())
        temp.close()
        return temp.name
    except:
        print("ERROR! Can't open url: %s" % url)
        return -1



def CreateBIDSParticipantFile(nidm_graph,output_file,participant_fields):
    '''
    Creates participant file based on requested fields

    :param nidm_graph:
    :param output_directory:
    :param fields:
    :return:
    '''

    print("Creating participants.json file...")
    fields = ["participant_id"]
    #fields.extend(participant_fields)
    participants=pd.DataFrame(columns=fields,index=[1])
    participants_json = {}

    #for each Constants.NIDM_SUBJECTID in NIDM file
    row_index=1
    for subj_uri,subj_id in nidm_graph.subject_objects(predicate=URIRef(Constants.NIDM_SUBJECTID.uri)):

        #adding subject ID to data list to append to participants data frame
        participants.loc[row_index,'participant_id',] = subj_id

        #for each of the fields in the participants list
        for fields in participant_fields:
            #if field identifier isn't a proper URI then do a fuzzy search on the graph, else an explicit search for the URL
            if(validators.url(fields)):
                #then this is a valid URI so simply query nidm_project document for it
                for subj,obj in nidm_graph.subject_objects(predicate=URIRef(BIDS_Constants.participants[fields].uri)):
                    #add row to the pandas data frame
                    #data.append(obj)
                    participants.loc[row_index,BIDS_Constants.participants[fields].uri] = obj
            else:
                #text matching task, remove basepart of URIs and try to fuzzy match the field in the part_fields parameter string
                #to the "term" part of a qname URI...this part let's a user simply ask for "age" for example without knowing the
                #complete URI....hopefully
                #
                #This needs to be a more complex query:
                #   Step(1): For subj_uri query for prov:Activity that were prov:wasAttributedTo subj_uri
                #   Step(2): Query for prov:Entity that were prov:wasGeneratedBy uris from Step(1)
                #   Step(3): For each metadata triple in objects whose subject is uris from Step(2), fuzzy match predicate after
                #   removing base of uri to "fields" in participants list, then add these to data list for appending to pandas
                match_ratio={}
                #
                #Steps(1):(3)

                query = """
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX prov: <http://www.w3.org/ns/prov#>
                    PREFIX onli: <http://neurolog.unice.fr/ontoneurolog/v3.0/instrument.owl#>
                    PREFIX sio: <http://semanticscience.org/ontology/sio.owl#>
                    PREFIX niiri: <http://iri.nidash.org/>

                SELECT DISTINCT ?pred ?value
                    WHERE {
                        ?asses_activity prov:qualifiedAssociation ?_blank .
    					?_blank rdf:type prov:Association ;
	                		prov:agent <%s> ;
                   			prov:hadRole sio:Subject .

                        ?entities prov:wasGeneratedBy ?asses_activity ;
                            rdf:type onli:assessment-instrument ;
                            ?pred ?value .
                        FILTER (regex(str(?pred) ,"%s","i" ))
                    }""" % (subj_uri,fields)
                # print(query)
                qres = nidm_graph.query(query)

                for row in qres:
                    #use last field in URIs for short column name and add full URI to sidecar participants.json file
                    url_parts = urllib.parse.urlsplit(row[0],scheme='#')

                    if url_parts.fragment == '':
                        #do some parsing of the path URL because this particular one has no fragments
                        url_parts = urllib.parse.urlparse(row[0])
                        path_parts = url_parts[2].rpartition('/')
                        short_name = path_parts[2]
                    else:
                        short_name = url_parts.fragment
                    participants_json[short_name] = {}
                    participants_json[short_name]['TermURL'] = row[0]

                    participants.loc[row_index,str(short_name)] = str(row[1])
                    #data.append(str(row[1]))

        #add row to participants DataFrame
        #participants=participants.append(pd.DataFrame(data))
        participants
        row_index = row_index+1


    #save participants.tsv file
    participants.to_csv(output_file + ".tsv",sep='\t',index=False)
    #save participants.json file
    with open(output_file + ".json",'w') as f:
        json.dump(participants_json,f,sort_keys=True,indent=2)

    return participants, participants_json



def NIDMProject2BIDSDatasetDescriptor(nidm_graph,output_directory):
    '''
    :param nidm_graph: RDFLib graph object from NIDM-Exp file
    :param output_dir: directory for writing dataset_description of BIDS dataset
    :return: None
    '''

    print("Creating dataset_description.json file...")

    #Project -> Dataset_description.json############################################
    #get json representation of project metadata
    project_metadata = nidm_graph.get_metadata_dict(Constants.NIDM_PROJECT)
    #print(project_metadata)

    #cycle through keys converting them to BIDS keys
    #make copy of project_metadata
    project_metadata_tmp = dict(project_metadata)
    #iterate over the temporary dictionary and delete items from the original
    for proj_key,value in project_metadata_tmp.items():
        key_found=0
        #print("proj_key = %s " % proj_key)
        #print("project_metadata[proj_key] = %s" %project_metadata[proj_key])

        for key,value in BIDS_Constants.dataset_description.items():
            if BIDS_Constants.dataset_description[key]._uri == proj_key:
                project_metadata[key] = project_metadata[proj_key]
                del project_metadata[proj_key]
                key_found=1
                continue
        #if this proj_key wasn't found in BIDS dataset_description Constants dictionary then delete it
        if not key_found:
            del project_metadata[proj_key]

    with open(join(output_directory, "dataset_description.json"),'w') as f:
        json.dump(project_metadata,f,sort_keys=True,indent=2)

    ##############################################################################

def ProcessFiles(graph,scan_type,output_directory,project_location,args):
    '''
    This function will essentially cycle through the acquisition objects in the NIDM file loaded into graph
    and depending on the scan_type will try and copy the image to the output_directory
    '''

    if scan_type == Constants.NIDM_MRI_DIFFUSION_TENSOR.uri:
        bids_ext = 'dwi'
    elif scan_type == Constants.NIDM_MRI_ANATOMIC_SCAN.uri:
        bids_ext = 'anat'
    elif scan_type == Constants.NIDM_MRI_FUNCTION_SCAN.uri:
        bids_ext = 'func'

    # query NIDM document for acquisition entity "subjects" with predicate nidm:hasImageUsageType and object scan_type
    for acq in graph.subjects(predicate=URIRef(Constants.NIDM_IMAGE_USAGE_TYPE.uri),
                                             object=URIRef(scan_type)):
        # first see if file exists locally.  Get nidm:Project prov:Location and append the nfo:Filename of the image
        # from the acq acquisition entity.  If that file doesn't exist try the prov:Location in the func acq
        # entity and see if we can download it from the cloud

        # get acquisition uuid from entity uuid
        temp = graph.objects(subject=acq, predicate=Constants.PROV['wasGeneratedBy'])
        for item in temp:
            activity = item
        # get participant ID with sio:Subject role in anat_acq qualified association
        part_id = GetParticipantIDFromAcquisition(nidm_file_list=[args.rdf_file], acquisition=activity)

        # make BIDS sub directory
        if 'sub' in (part_id['ID'].values)[0]:
            sub_dir = join(output_directory, (part_id['ID'].values)[0])
        else:
            sub_dir = join(output_directory, "sub-" + (part_id['ID'].values)[0])
        sub_filename_base = "sub-" + (part_id['ID'].values)[0]
        if not os.path.exists(sub_dir):
            os.makedirs(sub_dir)

        # make BIDS scan type directory (bids_ext) directory
        if not os.path.exists(join(sub_dir, bids_ext)):
            os.makedirs(join(sub_dir, bids_ext))

        for filename in graph.objects(subject=acq,predicate=URIRef(Constants.NIDM_FILENAME.uri)):
            # check if file exists
            for location in project_location:
                # if MRI exists in this location then copy and rename
                if isfile((location[0] + filename).lstrip("file:")):
                    # copy and rename file to be BIDS compliant
                    copyfile((location[0] + filename).lstrip("file:"),
                             join(sub_dir, bids_ext, sub_filename_base + splitext(filename)[1]))
                    continue
            # if the file wasn't accessible locally, try with the prov:Location in the acq
            for location in graph.objects(subject=acq,predicate=URIRef(Constants.PROV['Location'])):
                # try to download the file and rename
                ret = GetImageFromURL(location)
                if ret == -1:
                    print("ERROR! Can't download file: %s from url: %s, trying to copy locally...." % (
                    filename, location))
                    if "file" in location:
                        location = str(location).lstrip("file:")
                        print("Trying to copy file from %s" % (location))
                        try:
                            copyfile(location, join(output_directory, sub_dir, bids_ext, basename(filename)))
                        except:
                            print("ERROR! Failed to find file %s on filesystem..." % location)
                            if not args.no_downloads:
                                try:
                                    print(
                                        "Running datalad get command on dataset: %s" % location)
                                    dl.Dataset(os.path.dirname(location)).get(recursive=True, jobs=1)

                                except:
                                    print("ERROR! Datalad returned error: %s for dataset %s." % (
                                    sys.exc_info()[0], location))
                                    GetImageFromAWS(location=location, output_file=
                                        join(output_directory, sub_dir, bids_ext, basename(filename)),args=args)
                else:
                    # copy temporary file to BIDS directory
                    copyfile(ret, join(output_directory, sub_dir, bids_ext, basename(filename)))

            # if this is a DWI scan then we should copy over the b-value and b-vector files
            if bids_ext == 'dwi':
                # search for entity uuid with rdf:type nidm:b-value that was generated by activity
                query = """
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX prov: <http://www.w3.org/ns/prov#>
                    PREFIX nidm: <http://purl.org/nidash/nidm#>
    
                    SELECT DISTINCT ?entity
                        WHERE {
                            ?entity rdf:type <http://purl.org/nidash/nidm#b-value> ;
                                prov:wasGeneratedBy <%s> .
                        }""" % activity
                # print(query)
                qres = graph.query(query)

                for row in qres:
                    bval_entity = str(row[0])

                # if the file wasn't accessible locally, try with the prov:Location in the acq
                for location in graph.objects(subject=URIRef(bval_entity), predicate=URIRef(Constants.PROV['Location'])):
                    # try to download the file and rename
                    ret = GetImageFromURL(location)
                    if ret == -1:
                        print("ERROR! Can't download file: %s from url: %s, trying to copy locally...." % (
                            filename, location))
                        if "file" in location:
                            location = str(location).lstrip("file:")
                            print("Trying to copy file from %s" % (location))
                            try:
                                copyfile(location, join(output_directory, sub_dir, bids_ext, basename(location)))
                            except:
                                print("ERROR! Failed to find file %s on filesystem..." % location)
                                if not args.no_downloads:
                                    try:
                                        print(
                                            "Running datalad get command on dataset: %s" % location)
                                        dl.Dataset(os.path.dirname(location)).get(recursive=True, jobs=1)

                                    except:
                                        print("ERROR! Datalad returned error: %s for dataset %s." % (
                                            sys.exc_info()[0], location))
                                        GetImageFromAWS(location=location, output_file=
                                            join(output_directory, sub_dir, bids_ext, basename(location)),args=args)
                # search for entity uuid with rdf:type nidm:b-value that was generated by activity
                query = """
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX prov: <http://www.w3.org/ns/prov#>
                    PREFIX nidm: <http://purl.org/nidash/nidm#>

                    SELECT DISTINCT ?entity
                        WHERE {
                            ?entity rdf:type <http://purl.org/nidash/nidm#b-vector> ;
                                prov:wasGeneratedBy <%s> .
                        }""" % activity
                # print(query)
                qres = graph.query(query)

                for row in qres:
                    bvec_entity = str(row[0])

                # if the file wasn't accessible locally, try with the prov:Location in the acq
                for location in graph.objects(subject=URIRef(bvec_entity),
                                              predicate=URIRef(Constants.PROV['Location'])):
                    # try to download the file and rename
                    ret = GetImageFromURL(location)
                    if ret == -1:
                        print(
                            "ERROR! Can't download file: %s from url: %s, trying to copy locally...." % (
                                filename, location))
                        if "file" in location:
                            location = str(location).lstrip("file:")
                            print("Trying to copy file from %s" % (location))
                            try:
                                copyfile(location,
                                         join(output_directory, sub_dir, bids_ext, basename(location)))
                            except:
                                print("ERROR! Failed to find file %s on filesystem..." % location)
                                if not args.no_downloads:
                                    try:
                                        print(
                                            "Running datalad get command on dataset: %s" % location)
                                        dl.Dataset(os.path.dirname(location)).get(recursive=True,
                                                                                  jobs=1)

                                    except:
                                        print("ERROR! Datalad returned error: %s for dataset %s." % (
                                            sys.exc_info()[0], location))
                                        GetImageFromAWS(location=location, output_file=
                                            join(output_directory, sub_dir, bids_ext, basename(location)),
                                                        args=args)


def main(argv):
    parser = ArgumentParser(description='This program will convert a NIDM-Experiment RDF document \
        to a BIDS dataset.  The program will query the NIDM-Experiment document for subjects, \
        MRI scans, and associated assessments saving the MRI data to disk in an organization \
        according to the BIDS specification, metadata to a participants.tsv \
        file, the project-level metdata to a dataset_description.json file, and the \
        assessments to *.tsv/*.json file pairs in a phenotypes directory.', epilog='Example of use: \
        NIDM2BIDSMRI.py -nidm_file NIDM.ttl -part_fields age,gender -bids_dir BIDS')

    parser.add_argument('-nidm_file', dest='rdf_file', required=True, help="NIDM RDF file")
    parser.add_argument('-part_fields', nargs='+', dest='part_fields', required=False, \
                        help='Variables to add to BIDS participant file. Variables will be fuzzy-matched to NIDM URIs')
    parser.add_argument('-anat', dest='anat', action='store_true', required=False, help="Include flag to add anatomical scans to BIDS dataset")
    parser.add_argument('-func', dest='func', action='store_true', required=False, help="Include flag to add functional scans + events files to BIDS dataset")
    parser.add_argument('-dwi', dest='dwi', action='store_true', required=False, help="Include flag to add DWI scans + Bval/Bvec files to BIDS dataset")
    parser.add_argument('-bids_dir', dest='bids_dir', required=True, help="Directory to store BIDS dataset")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-no_downloads',dest='no_downloads', action='store_true',required=False, help=
                        "If this flag is set then script won't attempt to download images using datalad"
                        "and AWS S3.  Default behavior is files are downloaded if they don't exist locally.")
    group.add_argument('-aws_url', dest='aws_url', required=False, help="This tool facilites export of "
        "user-selected information from a NIDM file to a BIDS dataset and may have to fetch images. The NIDM files contain links from"
        "the local filesystem used to convert BIDS to NIDM and possibly DataLad dataset links to the files if the"
        " original BIDS data was a DataLad dataset. Here we support 3 modes of trying to find images: (1) copy from"
        " the local directory space using the prov:Location information in the NIDM file; (2) fetch the images from"
        " a DataLad remote if the original BIDS dataset was a DataLad dataset when bids2nidm was run; (3) attempt "
        " to download the images via a AWS S3 link.  This parameter lets the user set the base AWS S3 URL to try and"
        " find the images.  Currently it supports using the URL provided here and adding the dataset id, subject id,"
        " and filename.  For example, in OpenNeuro (OpenNeuro is supported by default but will serve as an example) the base AWS S3"
        " URL is \'s3://openneuro.org\'. The URL then becomes (for example) "
        " s3://openneuro.org/ds000002/sub-06/func/sub-06_task-probabilisticclassification_run-02_bold.nii.gz where this tool"
        " has added \'ds000002/sub-06/[FILENAME] to the base AWS S3 URL.")
    parser.add_argument('-dataset_string', dest='dataset_string', required=False, help="If -aws_url parameter is supplied"
        " this parameter (-dataset_string) is required as it will be added to the aws_baseurl to retrieve images for each"
        " subject and file.  For example, if -aws_baseurl is \'s3://davedata.org \' and -dataset_string is \'dataset1\' then"
        " the AWS S3 url for sub-1 and file sub1-task-rest_run-1_bold.nii.gz would be: "
        " \'s3://davedata.org/dataset1/sub-1/[anat | func | dwi/sub1-task-rest_run-1_bold.nii.gz\'")

    args = parser.parse_args()

    # check some argument dependencies
    if args.aws_url and not args.dataset_string:
        print("ERROR! You must include a -dataset_string if you supplied the -aws_baseurl.  If there is no dataset"
              " string in your AWS S3 urls then just supply -aws_baseurl with nothing after it.")
        print(args.print_help())
        exit(-1)

    # set up some local variables
    rdf_file = args.rdf_file
    output_directory = args.bids_dir

    # check if output directory exists, if not create it
    if not isdir(output_directory):
        mkdir(path=output_directory)


    #try to read RDF file
    print("Guessing RDF file format...")
    format_found=False
    for format in 'turtle','xml','n3','trix','rdfa':
        try:
            print("Reading RDF file as %s..." % format)
            #load NIDM graph into NIDM-Exp API objects
            nidm_project = read_nidm(rdf_file)
            print("RDF file sucessfully read")
            format_found=True
            break
        except Exception:
            print("File: %s appears to be an invalid %s RDF file" % (rdf_file,format))

    if not format_found:
        print("File doesn't appear to be a valid RDF format supported by Python RDFLib!  Please check input file")
        print("exiting...")
        exit(-1)

  #  if not os.path.isdir(join(output_directory,os.path.splitext(args.rdf_file)[0])):
  #      os.mkdir(join(output_directory,os.path.splitext(args.rdf_file)[0]))

    #convert Project NIDM object -> dataset_description.json file
    NIDMProject2BIDSDatasetDescriptor(nidm_project,output_directory)

    #create participants.tsv file.  In BIDS datasets there is no specification for how many or which type of assessment
    #variables might be in this file.  The specification does mention a minimum participant_id which indexes each of the
    #subjects in the BIDS dataset.
    #
    #if parameter -parts_field is defined then the variables listed will be fuzzy matched to the URIs in the NIDM file
    #and added to the participants.tsv file

    #use RDFLib here for temporary graph making query easier
    rdf_graph = Graph()
    rdf_graph_parse = rdf_graph.parse(source=StringIO(nidm_project.serializeTurtle()), format='turtle')

    # temporary write out turtle file for testing
    # rdf_graph_parse.serialize(destination="/Users/dbkeator/Downloads/ds000117.ttl", format='turtle')


    #create participants file
    CreateBIDSParticipantFile(rdf_graph_parse, join(output_directory, "participants"), args.part_fields)

    # get nidm:Project prov:Location
    # first get nidm:Project UUIDs
    project_uuid = GetProjectsUUID([rdf_file], output_file=None)
    project_location = []
    for uuid in project_uuid:
        project_location.append(GetProjectLocation(nidm_file_list=[rdf_file], project_uuid=uuid))

    #creating BIDS hierarchy with requested scans
    if args.anat==True:
        ProcessFiles(graph=rdf_graph_parse, scan_type=Constants.NIDM_MRI_ANATOMIC_SCAN.uri,
                     output_directory=output_directory, project_location=project_location, args=args)

    if args.func == True:
        ProcessFiles(graph=rdf_graph_parse, scan_type=Constants.NIDM_MRI_FUNCTION_SCAN.uri,
                     output_directory=output_directory, project_location=project_location, args=args)
    if args.dwi == True:
        ProcessFiles(graph=rdf_graph_parse, scan_type = Constants.NIDM_MRI_DIFFUSION_TENSOR.uri ,
                     output_directory=output_directory, project_location=project_location, args=args)

if __name__ == "__main__":
   main(sys.argv[1:])




import urllib
import re
import pprint

import pytest
import rdflib

from nidm.experiment import Project, Session, AssessmentAcquisition, AssessmentObject, Acquisition, AcquisitionObject, Query
from nidm.experiment.CDE import getCDEs
from nidm.core import Constants
from nidm.experiment.tools.rest import RestParser
import os
from pathlib import Path
from rdflib import Graph, util, URIRef

BRAIN_VOL_FILES = ['./cmu_a.nidm.ttl', './caltech.nidm.ttl']
OPENNEURO_FILES = ['ds000168.nidm.ttl']
ALL_FILES = ['./cmu_a.nidm.ttl', './caltech.nidm.ttl', 'ds000168.nidm.ttl']
OPENNEURO_PROJECT_URI = None
OPENNEURO_SUB_URI = None

test_person_uuid = ""
test_p2_subject_uuids = []
cmu_test_project_uuid = None
cmu_test_subject_uuid = None

@pytest.fixture(scope="module", autouse="True")
def setup():
    global cmu_test_project_uuid, cmu_test_subject_uuid, OPENNEURO_PROJECT_URI, OPENNEURO_SUB_URI

    if not Path('./cmu_a.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
            "cmu_a.nidm.ttl"
        )

    if not Path('./caltech.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/Caltech/nidm.ttl",
            "caltech.nidm.ttl"
        )

    restParser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    projects = restParser.run(BRAIN_VOL_FILES, '/projects')
    for p in projects:
        proj_info = restParser.run(BRAIN_VOL_FILES, '/projects/{}'.format(p))
        if 'dctypes:title' in proj_info.keys() and proj_info['dctypes:title'] == 'ABIDE CMU_a Site':
            cmu_test_project_uuid = p
            break

    subjects = restParser.run(BRAIN_VOL_FILES, '/projects/{}/subjects'.format(cmu_test_project_uuid))
    cmu_test_subject_uuid = subjects['uuid'][0]


    if not Path('./ds000168.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/openneuro/ds000168/nidm.ttl",
            "ds000168.nidm.ttl"
        )

    projects2 = restParser.run(OPENNEURO_FILES, '/projects')
    for p in projects2:
        proj_info = restParser.run(OPENNEURO_FILES, '/projects/{}'.format(p))
        if 'dctypes:title' in proj_info.keys() and proj_info['dctypes:title'] == 'Offline Processing in Associative Learning':
            OPENNEURO_PROJECT_URI = p
    subjects = restParser.run(OPENNEURO_FILES, '/projects/{}/subjects'.format(OPENNEURO_PROJECT_URI))
    OPENNEURO_SUB_URI = subjects['uuid'][0]



def test_rest_sub_id():

    restParser = RestParser()
    restParser.setOutputFormat(RestParser.OBJECT_FORMAT)

    result = restParser.run(ALL_FILES, '/projects/{}'.format(cmu_test_project_uuid))

    sub_id = result['subjects']['subject id'][5]
    sub_uuid = result['subjects']['uuid'][5]

    result2 = restParser.run(ALL_FILES, '/subjects/{}'.format(sub_id))

    pp = pprint.PrettyPrinter()
    pp.pprint('/subjects/{}'.format(sub_id))

    # make sure we got the same UUID when looking up by sub id
    assert result2['uuid'] == sub_uuid
    assert len(result2['instruments']) > 0



import urllib
import re

import pytest
import rdflib

from nidm.experiment import Project, Session, AssessmentAcquisition, AssessmentObject, Acquisition, AcquisitionObject, Query
from nidm.core import Constants
from nidm.experiment.tools.rest import RestParser
import os
from pathlib import Path
from rdflib import Graph, util, URIRef
import json


from prov.model import ProvAgent


REST_TEST_FILE = './agent.ttl'
BRAIN_VOL_FILES = ['./cmu_a.nidm.ttl', './caltech.nidm.ttl']
OPENNEURO_FILES = ['ds000001.nidm.ttl',
                   'ds000003.nidm.ttl',
                   'ds000011.nidm.ttl',
                   'ds000017.nidm.ttl',
                   'ds000101.nidm.ttl',
                   'ds000102.nidm.ttl',
                   'ds000113.nidm.ttl',
                   'ds000114.nidm.ttl',
                   'ds000120.nidm.ttl',
                   'ds000122.nidm.ttl',
                   'ds000138.nidm.ttl',
                   'ds000171.nidm.ttl',
                   'ds000208.nidm.ttl',
                   'ds000214.nidm.ttl',
                   'ds000220.nidm.ttl',
                   'ds000221.nidm.ttl',
                   'ds000224.nidm.ttl',
                   'ds000238.nidm.ttl',
                   'ds000245.nidm.ttl',
                   'ds000246.nidm.ttl',
                   'ds001021.nidm.ttl',
                   'ds001178.nidm.ttl',
                   'ds001226.nidm.ttl',
                   'ds001229.nidm.ttl',
                   'ds001232.nidm.ttl',
                   'ds001242.nidm.ttl'
                   ]

# OPENNEURO_FILES = ['ds000001.nidm.ttl',
#                    'ds000003.nidm.ttl']
#

ALL_FILES = ['./cmu_a.nidm.ttl', './caltech.nidm.ttl', 'ds000168.nidm.ttl']
OPENNEURO_PROJECT_URI = None
OPENNEURO_SUB_URI = None

test_person_uuid = ""
test_p2_subject_uuids = []
cmu_test_project_uuid = None
cmu_test_subject_uuid = None

@pytest.fixture(scope="module", autouse="True")
def setup():
    global cmu_test_project_uuid, cmu_test_subject_uuid, OPENNEURO_PROJECT_URI, OPENNEURO_SUB_URI



    for fname in OPENNEURO_FILES:
        dataset = fname.split('.')[0]
        if not Path('./{}'.format(fname)).is_file():
            urllib.request.urlretrieve(
                'https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/openneuro/{}/nidm.ttl'.format(dataset),
                fname
            )

    if not Path('./cmu_a.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
            "cmu_a.nidm.ttl"
        )

    if not Path('./caltech.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/Caltech/nidm.ttl",
            "caltech.nidm.ttl"
        )


def test_dataelement_list():
    rest_parser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    result = rest_parser.run(OPENNEURO_FILES, '/dataelements')

    assert type(result) == dict
    assert "data_elements" in result
    assert 'uuid' in result["data_elements"]
    assert 'label' in result["data_elements"]
    assert 'data_type_info' in result["data_elements"]

    assert len(result["data_elements"]["label"]) != 0
    assert len(result["data_elements"]["label"]) == len(result["data_elements"]["uuid"])
    assert len(result["data_elements"]["label"]) == len(result["data_elements"]["data_type_info"])

    for label in result["data_elements"]["label"]:
        assert label in [ str(x["label"]) for x in result["data_elements"]["data_type_info"] ]

    for uuid in result["data_elements"]["uuid"]:
        assert uuid in [ str(x["dataElementURI"]) for x in result["data_elements"]["data_type_info"] ]

    # now check for derivatives
    result = rest_parser.run(BRAIN_VOL_FILES, '/dataelements')
    assert type(result) == dict
    assert 'Left-WM-hypointensities Volume_mm3 (mm^3)' in result['data_elements']['label']



def test_dataelement_details():
    rest_parser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    # result = rest_parser.run(OPENNEURO_FILES, '/dataelements')
    #
    # dti = rest_parser.run(OPENNEURO_FILES, '/dataelements/{}'.format(result["data_elements"]["label"][0]))
    #
    # assert "label" in dti
    # assert "description" in dti
    # assert "isAbout" in dti
    # assert "inProjects" in dti
    #
    # # make sure the text formatter doesn't fail horribly
    # rest_parser.setOutputFormat(RestParser.CLI_FORMAT)
    # txt = rest_parser.run(OPENNEURO_FILES, '/dataelements/{}'.format(result["data_elements"]["label"][0]))
    #


    dti = rest_parser.run(OPENNEURO_FILES, '/dataelements/Left-WM-hypointensities Volume_mm3 (mm^3)')
    print (dti)



def test_dataelement_details_in_projects_field():
    rest_parser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    # result = rest_parser.run(OPENNEURO_FILES, '/dataelements')
    # dti = rest_parser.run(OPENNEURO_FILES, '/dataelements/{}'.format(result["data_elements"]["label"][0]))
    # assert len(dti['inProjects']) >= 1

    # find a data element that we are using for at least one subject
    data_element_label = 'Right-non-WM-hypointensities normMax (MR)'
    dti = rest_parser.run(BRAIN_VOL_FILES, '/dataelements/{}'.format(data_element_label))
    assert len (dti['inProjects']) >= 1



import tempfile
from urllib.request import urlretrieve
from nidm.core import Constants
import hashlib
from os import path, environ
import pickle
from rdflib import Graph
import nidm.experiment.Query




def download_cde_files():
    cde_dir = tempfile.gettempdir()

    for url in Constants.CDE_FILE_LOCATIONS:
        urlretrieve( url, "{}/{}".format(cde_dir, url.split('/')[-1] ) )

    return cde_dir


def getCDEs(file_list=None):

    if getCDEs.cache:
        return getCDEs.cache

    hasher = hashlib.md5()
    hasher.update(str(file_list).encode('utf-8'))
    h = hasher.hexdigest()

    cache_file_name = tempfile.gettempdir() + "/cde_graph.{}.pickle".format(h)

    if path.isfile(cache_file_name):
        rdf_graph = pickle.load(open(cache_file_name, "rb"))
        getCDEs.cache = rdf_graph
        return rdf_graph

    rdf_graph = Graph()

    if not file_list:

        cde_dir = ''
        if "CDE_DIR" in environ:
            cde_dir = environ['CDE_DIR']

        if (not cde_dir) and (path.isfile( '/opt/project/nidm/core/cde_dir/ants_cde.ttl' )):
            cde_dir = '/opt/project/nidm/core/cde_dir'

        if (not cde_dir):
            cde_dir = download_cde_files()

        file_list = [ ]
        for f in ['ants_cde.ttl', 'fs_cde.ttl', 'fsl_cde.ttl']:
            fname = '{}/{}'.format(cde_dir, f)
            if path.isfile( fname ):
                file_list.append( fname )



    for fname in file_list:
        if path.isfile(fname):
            cde_graph = nidm.experiment.Query.OpenGraph(fname)
            rdf_graph = rdf_graph + cde_graph




    cache_file = open(cache_file_name , 'wb')
    pickle.dump(rdf_graph, cache_file)
    cache_file.close()

    getCDEs.cache = rdf_graph
    return rdf_graph
getCDEs.cache = None


import os
import tempfile
import pandas as pd
import csv
from nidm.experiment.Query import GetProjectsUUID
import click
from nidm.experiment.tools.click_base import cli
from nidm.experiment.tools.rest import RestParser
import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import silhouette_score
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import LabelEncoder, MinMaxScaler
from sklearn import preprocessing

@cli.command()
@click.option("--nidm_file_list", "-nl", required=True,
              help="A comma separated list of NIDM files with full path")
@click.option("--var","-variables", required=True,
                 help="This parameter is for the variables the user would like to complete the k-means algorithm on.\nThe way this looks in the command is python3 nidm_kmeans.py -nl MTdemog_aseg_v2.ttl -v \"fs_003343,age*sex,sex,age,group,age*group,bmi\"")
@click.option("--k_range", "-k", required=True,
              help="The maxiumum number of clusters to try. The algorithm will go from 2 to this number to determine the optimal number of clusters.")
@click.option("--optimal_cluster_method", "-m", required=True,
              help="The criterion used to select the optimal partitioning (either Silhouette Score, AIC, or BIC).")
@click.option("--output_file", "-o", required=False,
              help="Optional output file (TXT) to store results of the linear regression, contrast, and regularization")
def gmm(nidm_file_list, output_file, var, k_range, optimal_cluster_method):
    """
            This function provides a tool to complete k-means clustering on NIDM data.
            """
    global v  # Needed to do this because the code only used the parameters in the first method, meaning I had to move it all to method 1.
    v = var.strip()  # used in data_aggregation, kmenas(), spaces stripped from left and right
    global o  # used in dataparsing()
    o = output_file
    global n  # used in data_aggregation()
    n = nidm_file_list
    global k_num
    k_num = int(k_range.strip())
    global cm
    cm = optimal_cluster_method
    data_aggregation()
    dataparsing()
    cluster_number()


def data_aggregation():  # all data from all the files is collected
    """    This function provides query support for NIDM graphs.   """
    # query result list
    results = []
    # if there is a CDE file list, seed the CDE cache
    if v:  #ex: age,sex,DX_GROUP
        print("***********************************************************************************************************")
        command = "pynidm k-means -nl " + n + " -variables \"" + v + "\" " + "-k " + str(k_num) + " -m " + cm

        print("Your command was: " + command)
        if (o is not None):
            f = open(o, "w")
            f.write("Your command was " + command)
            f.close()
        verbosity=0
        restParser = RestParser(verbosity_level=int(verbosity))
        restParser.setOutputFormat(RestParser.OBJECT_FORMAT)
        global df_list  # used in dataparsing()
        df_list = []
        # set up uri to do fields query for each nidm file
        global file_list
        file_list = n.split(",")
        df_list_holder = {}
        for i in range(len(file_list)):
            df_list_holder[i] = []
        df_holder = {}
        for i in range(len(file_list)):
            df_holder[i] = []
        global condensed_data_holder
        condensed_data_holder = {}
        for i in range(len(file_list)):
            condensed_data_holder[i] = []

        count = 0
        not_found_count = 0
        for nidm_file in file_list:
            # get project UUID
            project = GetProjectsUUID([nidm_file])
            # split the model into its constituent variables
            global var_list
            # below, we edit the model so it splits by +,~, or =. However, to help it out in catching everything
            # we replaced ~ and = with a + so that we can still use split. Regex wasn't working.
            var_list = v.split(",")
            for i in range(len(var_list)):  # here, we remove any leading or trailing spaces
                var_list[i] = var_list[i].strip()
            # set the dependent variable to the one dependent variable in the model
            global vars  # used in dataparsing()
            vars = ""
            for i in range(len(var_list) - 1, -1, -1):
                if not "*" in var_list[i]:  # removing the star term from the columns we're about to pull from data
                    vars = vars + var_list[i] + ","
                else:
                    print("Interacting variables are not present in clustering models. They will be removed.")
            vars = vars[0:len(vars) - 1]
            uri = "/projects/" + project[0].toPython().split("/")[-1] + "?fields=" + vars
            # get fields output from each file and concatenate
            df_list_holder[count].append(pd.DataFrame(restParser.run([nidm_file], uri)))
            # global dep_var
            df = pd.concat(df_list_holder[count])
            with tempfile.NamedTemporaryFile(delete=False) as temp:  # turns the dataframe into a temporary csv
                df.to_csv(temp.name + '.csv')
                temp.close()
            data = list(csv.reader(open(
                temp.name + '.csv')))  # makes the csv a 2D list to make it easier to call the contents of certain cells

            var_list = vars.split(",")  # makes a list of the independent variables
            numcols = (len(data) - 1) // (
                    len(var_list))  # Finds the number of columns in the original dataframe
            global condensed_data  # also used in linreg()
            condensed_data_holder[count] = [
                [0] * (len(var_list))]  # makes an array 1 row by the number of necessary columns
            for i in range(
                    numcols):  # makes the 2D array big enough to store all of the necessary values in the edited dataset
                condensed_data_holder[count].append([0] * (len(var_list)))
            for m in range(0, len(var_list)):
                end_url = var_list[m].split("/")
                if "/" in var_list[m]:
                    var_list[m] = end_url[len(end_url) - 1]
            for i in range(len(var_list)):  # stores the independent variable names in the first row
                condensed_data_holder[count][0][i] = var_list[i]
            numrows = 1  # begins at the first row to add data
            fieldcolumn = 0  # the column the variable name is in in the original dataset
            valuecolumn = 0  # the column the value is in in the original dataset
            datacolumn = 0  # if it is identified by the dataElement name instead of the field's name
            not_found_list = []
            for i in range(len(data[0])):
                if data[0][i] == 'sourceVariable':  # finds the column where the variable names are
                    fieldcolumn = i
                elif data[0][i] == 'source_variable':  # finds the column where the variable names are
                    fieldcolumn = i
                elif data[0][i] == 'isAbout':
                    aboutcolumn = i
                elif data[0][i] == 'label':
                    namecolumn = i  # finds the column where the variable names are
                elif data[0][i] == 'value':
                    valuecolumn = i  # finds the column where the values are
                elif data[0][i] == 'dataElement':  # finds the column where the data element is if necessary
                    datacolumn = i
            for i in range(
                    len(condensed_data_holder[count][
                            0])):  # starts iterating through the dataset, looking for the name in that
                for j in range(1, len(data)):  # column, so it can append the values under the proper variables
                    try:
                        if data[j][fieldcolumn] == condensed_data_holder[count][0][
                            i]:  # in the dataframe, the name is in column 3
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif data[j][aboutcolumn] == condensed_data_holder[count][0][
                            i]:
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif condensed_data_holder[count][0][
                            i] in data[j][
                            aboutcolumn]:  # this is in case the uri only works by querying the part after the last backslash
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif data[j][namecolumn] == condensed_data_holder[count][0][
                            i]:  # in the dataframe, the name is in column 12
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif condensed_data_holder[count][0][i] == data[j][
                            datacolumn]:  # in the dataframe, the name is in column 9
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                    except IndexError:
                        numrows = numrows + 1
                numrows = 1  # resets to the first row for the next variable
            temp_list = condensed_data_holder[count]
            for j in range(len(temp_list[0]) - 1, 0,
                           -1):  # if the software appends a column with 0 as the heading, it removes this null column
                if temp_list[0][j] == "0" or temp_list[0][j] == "NaN":
                    for row in condensed_data_holder[count]:
                        row.pop(j)
            rowsize = len(condensed_data_holder[count][0])
            count1 = 0
            for i in range(0, rowsize):
                for row in condensed_data_holder[count]:
                    if row[i] == 0 or row[i] == "NaN" or row[i] == "0":
                        count1 = count1 + 1
                if count1 > len(condensed_data_holder[count]) - 2:
                    not_found_list.append(condensed_data_holder[count][0][i])
                count1 = 0
            for i in range(len(condensed_data_holder[count][0])):
                if " " in condensed_data_holder[count][0][i]:
                    condensed_data_holder[count][0][i] = condensed_data_holder[count][0][i].replace(" ", "_")
            for i in range(len(var_list)):
                if "/" in var_list[i]:
                    splitted = var_list[i].split("/")
                    var_list[i] = splitted[len(splitted) - 1]
                if " " in var_list[i]:
                    var_list[i] = var_list[i].replace(" ", "_")
            count = count + 1
            if len(not_found_list) > 0:
                print(
                    "***********************************************************************************************************")
                print()
                print("Your variables were " + v)
                print()
                print(
                    "The following variables were not found in " + nidm_file + ". The model cannot run because this will skew the data. Try checking your spelling or use nidm_query.py to see other possible variables.")
                if (o is not None):
                    f = open(o, "a")
                    f.write("Your variables were " + v)
                    f.write(
                        "The following variables were not found in " + nidm_file + ". The model cannot run because this will skew the data. Try checking your spelling or use nidm_query.py to see other possible variables.")
                    f.close()
                for i in range(0, len(not_found_list)):
                    print(str(i + 1) + ". " + not_found_list[i])
                    if (o is not None):
                        f = open(o, "a")
                        f.write(str(i + 1) + ". " + not_found_list[i])
                        f.close()
                for j in range(len(not_found_list) - 1, 0, -1):
                    not_found_list.pop(j)
                not_found_count = not_found_count + 1
                print()
        if not_found_count > 0:
            exit(1)


    else:
        print("ERROR: No query parameter provided.  See help:")
        print()
        os.system("pynidm k-means --help")
        exit(1)

def dataparsing(): #The data is changed to a format that is usable by the linear regression method
    global condensed_data
    condensed_data = []
    for i in range(0, len(file_list)):
        condensed_data = condensed_data + condensed_data_holder[i]
    global k_num
    if len(condensed_data[0])<=k_num:
        print("\nThe maximum number of clusters specified is greater than the amount of data present.")
        print("The algorithm cannot run with this, so k_num will be reduced to 1 less than the length of the dataset.")
        k_num = len(condensed_data) -1
        print("The k_num value is now: " + str(k_num))
    x = pd.read_csv(opencsv(condensed_data))  # changes the dataframe to a csv to make it easier to work with
    x.head()  # prints what the csv looks like
    x.dtypes  # checks data format
    obj_df = x.select_dtypes  # puts all the variables in a dataset
    x.shape  # says number of rows and columns in form of tuple
    x.describe()  # says dataset statistics
    obj_df = x.select_dtypes(
        include=['object']).copy()  # takes everything that is an object (not float or int) and puts it in a new dataset
    obj_df.head()  # prints the new dataset
    int_df = x.select_dtypes(include=['int64']).copy()  # takes everything that is an int and puts it in a new dataset
    float_df = x.select_dtypes(
        include=['float64']).copy()  # takes everything that is a float and puts it in a new dataset
    df_int_float = pd.concat([float_df, int_df], axis=1)
    stringvars = []  # starts a list that will store all variables that are not numbers
    for i in range(1, len(condensed_data)):  # goes through each variable
        for j in range(len(condensed_data[0])):  # in the 2D array
            try:  # if the value of the field can be turned into a float (is numerical)
                float(condensed_data[i][j])  # this means it's a number
            except ValueError:  # if it can't be (is a string)
                if condensed_data[0][
                    j] not in stringvars:  # adds the variable name to the list if it isn't there already
                    stringvars.append(condensed_data[0][j])
    le = preprocessing.LabelEncoder()  # anything involving le shows the encoding of categorical variables
    for i in range(len(stringvars)):
        le.fit(obj_df[stringvars[i]].astype(str))
    obj_df_trf = obj_df.astype(str).apply(le.fit_transform)  # transforms the categorical variables into numbers.
    global df_final  # also used in linreg()
    if not obj_df_trf.empty:
        df_final = pd.concat([df_int_float, obj_df_trf], axis=1)  # join_axes=[df_int_float.index])
    else:
        df_final = df_int_float
    df_final.head()  # shows the final dataset with all the encoding
    print(df_final)  # prints the final dataset
    print()
    print("***********************************************************************************************************")
    print()
    if (o is not None):
        f = open(o, "a")
        f.write(df_final.to_string(header=True, index=True))
        f.write(
            "\n\n***********************************************************************************************************")
        f.write("\n\nModel Results: ")
        f.close()
def cluster_number():
    index = 0
    global levels  # also used in contrasting()
    levels = []
    for i in range(1, len(condensed_data)):
        if condensed_data[i][index] not in levels:
            levels.append(condensed_data[i][index])
    for i in range(len(levels)):
        levels[i] = i

    # Beginning of the linear regression
    global X
    #global y
    #Unsure on how to procede here with interacting variables, since I'm sure dmatrices won't work

    """scaler = MinMaxScaler()

    for i in range(len(model_list)):
        scaler.fit(df_final[[model_list[i]]])
        df_final[[model_list[i]]] = scaler.transform(df_final[[model_list[i]]])"""
    X = df_final[var_list]
    if "si" in cm.lower():
        print("Sillhoute Score")

        ss = []

        for i in range(2, k_num):
            model = GaussianMixture(n_components=i, init_params='kmeans')
            cluster_labels = model.fit_predict(X)
            silhouette_avg = silhouette_score(X, cluster_labels)
            ss.append(silhouette_avg)
        optimal_i = 0
        distance_to_one = abs(1-ss[0])
        for i in range(0,len(ss)):
            if abs(1-ss[i]) <= distance_to_one:
                optimal_i = i
                distance_to_one = abs(1-ss[i])

        n_clusters = optimal_i + 2
        print("Optimal number of clusters: " + str(n_clusters)) #optimal number of clusters
        gmm = GaussianMixture(n_components=n_clusters).fit(X)
        labels = gmm.fit(X).predict(X)
        ax = None or plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap='viridis', zorder=2)
        ax.axis('equal')
        plt.show()

    if "a" in cm.lower():
        print("AIC\n")
        aic = []
        for i in range(2, k_num):
            model = GaussianMixture(n_components=i, init_params='kmeans')
            model.fit(X)
            aic.append(model.bic(X))
        min_aic = aic[0]
        min_i = 0
        for i in range(1, len(aic)):
            if aic[i] <= min_aic:
                min_aic = aic[i]
                min_i = i
        n_clusters = min_i +2
        print("Optimal number of clusters: " + str(n_clusters)) #optimal number of clusters, minimizing aic
        """min_aic = aic[0]
        max_aic = aic[0]
        max_i = 0
        min_i = 0
        for i in range(1, len(aic)):
            if aic[i] >= max_aic:
                max_aic = aic[i]
                max_i = i
            elif aic[i] <= min_aic:
                min_aic = aic[i]
                min_i = i
        p1 = np.array([min_i, aic[min_i]])
        p2 = np.array([max_i, aic[max_i]])
        # the way I am doing the method is as follows:
        # the different sse values form a curve like an L (like an exponential decay)
        # The elbow is the point furthest from a line connecting max and min
        # So I am calculating the distance, and the maximum distance from point to curve shows the optimal point
        # AKA the number of clusters
        dist = []
        for n in range(0, len(aic)):
            norm = np.linalg.norm
            p3 = np.array([n, aic[n]])
            dist.append(np.abs(norm(np.cross(p2 - p1, p1 - p3))) / norm(p2 - p1))
        max_dist = dist[0]
        n_clusters = 2
        for x in range(1, len(dist)):
            if dist[x] >= max_dist:
                max_dist = dist[x]
                n_clusters = x + 2

        plt.plot(aic)
        plt.show()"""

        gmm = GaussianMixture(n_components=n_clusters).fit(X)
        labels = gmm.fit(X).predict(X)
        ax = None or plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap='viridis', zorder=2)
        ax.axis('equal')
        plt.show()

    if "b" in cm.lower():
        print("\n\nBIC\n")
        bic = []
        for i in range(2, k_num):
            model = GaussianMixture(n_components=i, init_params='kmeans')
            model.fit(X)
            bic.append(model.bic(X))
        min_bic = bic[0]
        min_i = 0
        for i in range(1, len(bic)):
            if bic[i] <= min_bic:
                min_bic = bic[i]
                min_i = i
        n_clusters = min_i + 2
        """min_bic = bic[0]
        max_bic = bic[0]
        max_i = 0
        min_i = 0
        for i in range(1,len(bic)):
            if bic[i]>=max_bic:
                max_bic = bic[i]
                max_i = i
            elif bic[i]<= min_bic:
                min_bic = bic[i]
                min_i = i
        p1 = np.array([min_i, bic[min_i]])
        p2 = np.array([max_i, bic[max_i]])
        # the way I am doing the method is as follows:
        # the different sse values form a curve like an L (like an exponential decay)
        # The elbow is the point furthest from a line connecting max and min
        # So I am calculating the distance, and the maximum distance from point to curve shows the optimal point
        # AKA the number of clusters
        dist = []
        for n in range(0, len(bic)):
            norm = np.linalg.norm
            p3 = np.array([n, bic[n]])
            dist.append(np.abs(norm(np.cross(p2 - p1, p1 - p3))) / norm(p2 - p1))
        max_dist = dist[0]
        n_clusters = 2
        for x in range(1, len(dist)):
            if dist[x] >= max_dist:
                max_dist = dist[x]
                n_clusters = x + 2
        plt.plot(bic)
        plt.show()"""
        print("Optimal number of clusters: " + str(n_clusters)) #optimal number of clusters
        gmm = GaussianMixture(n_components=n_clusters).fit(X)
        labels = gmm.fit(X).predict(X)
        ax = None or plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap='viridis', zorder=2)
        ax.axis('equal')
        plt.show()

    if (o is not None):
        f = open(o, "a")
        f.close()
def opencsv(data):
    """saves a list of lists as a csv and opens"""
    import tempfile
    import os
    import csv
    handle, fn = tempfile.mkstemp(suffix='.csv')
    with os.fdopen(handle,"w", encoding='utf8',errors='surrogateescape',newline='') as f:
        writer = csv.writer(f)
        writer.writerows(data)
    return fn
# it can be used calling the script `python nidm_query.py -nl ... -q ..
if __name__ == "__main__":
    gmm()




from pathlib import Path
import pytest
import pandas as pd
import json
import os
import urllib
import re
from  nidm.experiment.Utils import map_variables_to_terms
import tempfile
from os.path import join
from nidm.core import Constants
from uuid import UUID





@pytest.fixture(scope="module", autouse="True")
def setup():
    global DATA, REPROSCHEMA_JSON_MAP, BIDS_SIDECAR

    temp = { 'participant_id': ['100', '101', '102', '103', '104', '105', '106', '107', '108', '109'],
             'age': [18, 25, 30,19 ,35 ,20 ,27 ,29 ,38 ,27],
             'sex': ['m', 'm', 'f', 'm', 'f', 'f', 'f', 'f', 'm','m'] }

    DATA = pd.DataFrame(temp)

    REPROSCHEMA_JSON_MAP = json.loads(
        '''
        {
            "DD(source='participants.tsv', variable='participant_id')": {
                "label": "participant_id",
                "description": "subject/participant identifier",
                "source_variable": "participant_id",
                "responseOptions": {
                    "valueType": "http://www.w3.org/2001/XMLSchema#string"
                },
                "isAbout": [
                    {
                        "@id": "https://ndar.nih.gov/api/datadictionary/v2/dataelement/src_subject_id",
                        "label": "src_subject_id"
                    }
                ]
            },
            "DD(source='participants.tsv', variable='age')": {
                "responseOptions": {
                    "unitCode": "years",
                    "minValue": "0",
                    "maxValue": "100",
                    "valueType": "http://www.w3.org/2001/XMLSchema#integer"
                },
                "label": "age",
                "description": "age of participant",
                "source_variable": "age",
                "associatedWith": "NIDM",
                "isAbout": [
                    {
                        "@id": "http://uri.interlex.org/ilx_0100400",
                        "label": "Age"
                    }
                ]
            },
            "DD(source='participants.tsv', variable='sex')": {
                "responseOptions": {
                    "minValue": "NA",
                    "maxValue": "NA",
                    "unitCode": "NA",
                    "valueType": "http://www.w3.org/2001/XMLSchema#complexType",
                    "choices": {
                        "Male": "m",
                        "Female": "f"
                    }
                },
                "label": "sex",
                "description": "biological sex of participant",
                "source_variable": "sex",
                "associatedWith": "NIDM",
                "isAbout": [
                    {
                        "@id": "http://uri.interlex.org/ilx_0738439",
                        "label": "SEX"
                    }
                ]
            }
        }''')

    BIDS_SIDECAR = json.loads(
        '''
        {
            "age": {
                "label": "age",
                "description": "age of participant",
                "source_variable": "age",
                "associatedWith": "NIDM",
                "isAbout": [
                    {
                        "@id": "http://uri.interlex.org/ilx_0100400",
                        "label": "Age"
                    }
                ],
                "valueType": "http://www.w3.org/2001/XMLSchema#integer",
                "minValue": "10",
                "maxValue": "100"
            },
            "sex": {
                "minValue": "NA",
                "maxValue": "NA",
                "unitCode": "NA",
                "valueType": "http://www.w3.org/2001/XMLSchema#complexType",
                "levels": {
                    "Male": "m",
                    "Female": "f"
                },
                "label": "sex",
                "description": "biological sex of participant",
                "source_variable": "sex",
                "associatedWith": "NIDM",
                "isAbout": [
                    {
                        "@id": "http://uri.interlex.org/ilx_0738439",
                        "label": "SEX"
                    }
                ]
            }
        }
        
        ''')


def test_map_vars_to_terms_BIDS():
    '''
    This function will test the Utils.py "map_vars_to_terms" function with a BIDS-formatted
    JSON sidecar file
    '''


    global DATA, BIDS_SIDECAR

    column_to_terms, cde = map_variables_to_terms(df=DATA,json_source=BIDS_SIDECAR,
                directory=tempfile.gettempdir(),assessment_name="test",bids=True)

    # check whether JSON mapping structure returned from map_variables_to_terms matches the
    # reproshema structure
    assert "DD(source='test', variable='age')" in column_to_terms.keys()
    assert "DD(source='test', variable='sex')" in column_to_terms.keys()
    assert "isAbout" in column_to_terms["DD(source='test', variable='age')"].keys()
    assert "http://uri.interlex.org/ilx_0100400" == column_to_terms["DD(source='test', variable='age')"] \
            ['isAbout'][0]['@id']
    assert "http://uri.interlex.org/ilx_0738439" == column_to_terms["DD(source='test', variable='sex')"] \
                ['isAbout'][0]['@id']
    assert "responseOptions" in column_to_terms["DD(source='test', variable='sex')"].keys()
    assert "choices" in column_to_terms["DD(source='test', variable='sex')"]['responseOptions'].keys()
    assert "Male" in column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices'].keys()
    assert "m" == column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices']['Male']
    assert "Male" in column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices'].keys()
    assert "m" == column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices']['Male']

    # now check the JSON sidecar file created by map_variables_to_terms which should match BIDS format
    with open(join(tempfile.gettempdir(),"nidm_annotations.json")) as fp:
        bids_sidecar = json.load(fp)

    assert "age" in bids_sidecar.keys()
    assert "sex" in bids_sidecar.keys()
    assert "isAbout" in bids_sidecar["age"].keys()
    assert "http://uri.interlex.org/ilx_0100400" == bids_sidecar["age"] \
        ['isAbout'][0]['@id']
    assert "http://uri.interlex.org/ilx_0738439" == bids_sidecar["sex"] \
        ['isAbout'][0]['@id']
    assert "levels" in bids_sidecar["sex"].keys()
    assert "Male" in bids_sidecar["sex"]['levels'].keys()
    assert "m" == bids_sidecar["sex"]['levels']['Male']
    assert "Male" in bids_sidecar["sex"]['levels'].keys()
    assert "m" == bids_sidecar["sex"]['levels']['Male']

    # check the CDE dataelement graph for correct information
    query = '''
        prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        
        select distinct ?uuid ?DataElements ?property ?value
            where {

                ?uuid a/rdfs:subClassOf* nidm:DataElement ;
                    ?property ?value .

        }'''
    qres=cde.query(query)

    results=[]
    for row in qres:
        results.append(list(row))

    assert len(results) == 20


def test_map_vars_to_terms_reproschema():
    '''
    This function will test the Utils.py "map_vars_to_terms" function with a reproschema-formatted
    JSON sidecar file
    '''

    global DATA, REPROSCHEMA_JSON_MAP

    column_to_terms, cde = map_variables_to_terms(df=DATA, json_source=REPROSCHEMA_JSON_MAP,
                                                  directory=tempfile.gettempdir(), assessment_name="test")

    # check whether JSON mapping structure returned from map_variables_to_terms matches the
    # reproshema structure
    assert "DD(source='test', variable='age')" in column_to_terms.keys()
    assert "DD(source='test', variable='sex')" in column_to_terms.keys()
    assert "isAbout" in column_to_terms["DD(source='test', variable='age')"].keys()
    assert "http://uri.interlex.org/ilx_0100400" == column_to_terms["DD(source='test', variable='age')"] \
        ['isAbout'][0]['@id']
    assert "http://uri.interlex.org/ilx_0738439" == column_to_terms["DD(source='test', variable='sex')"] \
        ['isAbout'][0]['@id']
    assert "responseOptions" in column_to_terms["DD(source='test', variable='sex')"].keys()
    assert "choices" in column_to_terms["DD(source='test', variable='sex')"]['responseOptions'].keys()
    assert "Male" in column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices'].keys()
    assert "m" == column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices']['Male']
    assert "Male" in column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices'].keys()
    assert "m" == column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices']['Male']

    # now check the JSON mapping file created by map_variables_to_terms which should match Reproschema format
    with open(join(tempfile.gettempdir(), "nidm_annotations.json")) as fp:
        reproschema_json = json.load(fp)

    assert "DD(source='test', variable='age')" in column_to_terms.keys()
    assert "DD(source='test', variable='sex')" in column_to_terms.keys()
    assert "isAbout" in column_to_terms["DD(source='test', variable='age')"].keys()
    assert "http://uri.interlex.org/ilx_0100400" == column_to_terms["DD(source='test', variable='age')"] \
        ['isAbout'][0]['@id']
    assert "http://uri.interlex.org/ilx_0738439" == column_to_terms["DD(source='test', variable='sex')"] \
        ['isAbout'][0]['@id']
    assert "responseOptions" in column_to_terms["DD(source='test', variable='sex')"].keys()
    assert "choices" in column_to_terms["DD(source='test', variable='sex')"]['responseOptions'].keys()
    assert "Male" in column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices'].keys()
    assert "m" == column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices']['Male']
    assert "Male" in column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices'].keys()
    assert "m" == column_to_terms["DD(source='test', variable='sex')"]['responseOptions']['choices']['Male']

    # check the CDE dataelement graph for correct information
    query = '''
        prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>

        select distinct ?uuid ?DataElements ?property ?value
            where {

                ?uuid a/rdfs:subClassOf* nidm:DataElement ;
                    ?property ?value .

        }'''
    qres = cde.query(query)

    results = []
    for row in qres:
        results.append(list(row))

    assert len(results) == 20


    

